import pathlib

import pytest

from moo.core import code, create, lookup, parse
from moo.core.models import Object
from moo.core.models.verb import Repository, Verb, VerbName
from moo.bootstrap import load_verb_source


def _write_verb_file(path: pathlib.Path, on: str, verb_name: str, body: str) -> pathlib.Path:
    """Write a minimal verb source file to *path* and return it."""
    path.write_text(f"#!moo verb {verb_name} --on {on}\n{body}\n", encoding="utf8")
    return path


# --- @reload ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_reload_updates_verb_code(t_init: Object, t_wizard: Object, tmp_path):
    """@reload updates verb code in the DB when the source file has changed."""
    system = lookup(1)
    repo = Repository.objects.create(slug="test-repo", prefix="test", url="http://example.com")

    with code.ContextManager(t_wizard, lambda _: None) as ctx:
        obj = create("widget", parents=[system.root_class], location=t_wizard.location)

        verb_file = tmp_path / "myverb.py"
        _write_verb_file(verb_file, "widget", "myverb", 'print("original")')
        load_verb_source(verb_file, system, repo)

        assert obj.has_verb("myverb")
        assert "original" in obj.get_verb("myverb").code

        # Modify the file on disk then reload via the verb
        _write_verb_file(verb_file, "widget", "myverb", 'print("updated")')
        parse.interpret(ctx, "@reload myverb on widget")

    obj.refresh_from_db()
    assert "updated" in obj.get_verb("myverb").code


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_reload_nonexistent_verb_no_output(t_init: Object, t_wizard: Object, tmp_path):
    """@reload on a verb that doesn't exist produces no output and no exception."""
    system = lookup(1)
    repo = Repository.objects.create(slug="test-repo", prefix="test", url="http://example.com")

    printed = []
    with code.ContextManager(t_wizard, printed.append) as ctx:
        create("widget", parents=[system.root_class], location=t_wizard.location)

        verb_file = tmp_path / "myverb.py"
        _write_verb_file(verb_file, "widget", "myverb", 'print("hello")')
        load_verb_source(verb_file, system, repo)

        parse.interpret(ctx, "@reload bogusverb on widget")
    assert printed == []


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_reload_verb_without_filename_raises(t_init: Object, t_wizard: Object):
    """@reload on a verb with no filename/repo raises RuntimeError."""
    system = lookup(1)
    with code.ContextManager(t_wizard, lambda _: None) as ctx:
        obj = create("widget", parents=[system.root_class], location=t_wizard.location)
        v = Verb.objects.create(origin=obj, owner=t_wizard, code='print("hi")')
        VerbName.objects.create(verb=v, name="myverb")
        with pytest.raises(RuntimeError):
            parse.interpret(ctx, "@reload myverb on widget")



# -*- coding: utf-8 -*-
"""
Useful global utilities.
"""

import warnings


def apply_default_permissions(instance):
    from .models import Object
    from .models.verb import Verb
    system = Object.objects.get(pk=1)
    set_default_permissions = Verb.objects.filter(
        origin=system, names__name="set_default_permissions"
    )
    set_default_permissions = set_default_permissions.first()
    if set_default_permissions:
        set_default_permissions.invoked_name = "set_default_permissions"
        set_default_permissions.invoked_object = system
        set_default_permissions(instance)
    else:
        warnings.warn(f"set_default_permissions failed for {instance}: verb not found", category=RuntimeWarning)

def expand_wildcard(name):
    if "*" not in name:
        return [name]
    prefix, suffix = name.split("*", 1)
    result = [prefix]
    for i in range(len(suffix)):
        result.append(prefix + suffix[:i+1])
    return result



#!moo verb huh --on $room

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is a stub used to call the huh2 verb. It is called by the LambdaMOO server when it can't match a sentence
given to it by a player. The server calls `huh` with verb equal to the actual verb in the erroneous command line.
This means it is not possible to use passthrough() if you override `huh` in a room; it would pass control up to a verb on
the parent named verb, i.e., whatever the verb was on the command line. This, by definition, doesn't exist. To get
around this, and allow you to use passthrough() to get the default `huh` behaviour, `huh` calls `huh2`. You should override
`huh2` if you wish to be able to use passthrough() to get the default `huh` behaviour.
"""

from moo.core import context

return this.huh2(context.parser.words[0])



#!moo verb look --on "room class"

# pylint: disable=return-outside-function,undefined-variable

qs = this.properties.filter(name="description")  # pylint: disable=undefined-variable. # type: ignore
if qs:
    print(qs[0].value)
else:
    print("No description.")



"""
Encode/decode MOO JSON
"""

import json


def loads(j):
    from .models.object import Object
    from .models.property import Property
    from .models.verb import Verb

    def to_entity(d):
        if len(d) != 1:
            return d
        key = list(d.keys())[0]
        if key[1] == "#":
            if key[0] == "o":
                return Object.objects.get(pk=int(key[2:]))
            elif key[0] == "v":
                return Verb.objects.get(pk=int(key[2:]))
            elif key[0] == "p":
                return Property.objects.get(pk=int(key[2:]))
        return d

    return json.loads(j, object_hook=to_entity)


def dumps(obj):
    from .models.object import Object
    from .models.property import Property
    from .models.verb import Verb

    def from_entity(o):
        if isinstance(o, Object):
            return {"o#%d" % o.pk: o.name}
        elif isinstance(o, Verb):
            return {"v#%d" % o.pk: o.name()}
        elif isinstance(o, Property):
            return {"p#%d" % o.pk: o.name}
        else:
            raise TypeError(
                "Unserializable object {} of type {}".format(obj, type(obj))
            )
    return json.dumps(obj, default=from_entity)



#!moo verb @dig @tunnel --on $player --dspec any --ispec "through:any" --ispec "to:any"

# pylint: disable=return-outside-function,undefined-variable

"""
This is a player command used to create a room or exit, (that is, instances of the class `$room` or `$exit'. The verb
parses the arguments to determine the type and number of objects that are to be created. It uses the `create()` primitive,
with `$room` as a parent to create a room. Note that you can only use the `@dig` command to dig an exit from within a
room.
"""

from moo.core import context, create, lookup

directions = ["north", "northeast", "east", "southeast", "south", "southwest", "west", "northwest", "up", "down"]
direction = context.parser.get_dobj_str()

source = context.player.location
if source.match_exit(direction):
    print("[color red]There is already an exit in that direction.[/color red]")
    return

if context.parser.has_pobj("through"):
    door = context.parser.get_pobj("through")
    if not door.is_a(_.exit):
        print("[color red]The specified object is not an exit.[/color red]")
        return
    if not door.aliases.filter(alias=direction):
        door.aliases.create(alias=direction)
else:
    door = create(f"{direction} from {source.name}", parents=[_.exit], location=None)
    door.aliases.create(alias=direction)

if verb_name == "@dig":
    action = "Dug"
    dest = create(context.parser.get_pobj_str("to"), parents=[_.room], location=None)
else:
    action = "Tunnelled"
    dest = context.parser.get_pobj("to", lookup=True)

door.set_property("source", source)
door.set_property("dest", dest)
source.add_exit(door)
dest.add_entrance(door)

print(f'[color yellow]{action} an exit {direction} to "{dest.name}".[/color yellow]')



# -*- coding: utf-8 -*-
"""
Celery configuration for the moo project.
https://docs.celeryq.dev/en/stable/userguide/configuration.html
"""

import os

redis_hostname = os.environ.get("REDIS_HOSTNAME", "redis")
broker_url = os.environ.get("CELERY_BROKER_URL", f"redis://{redis_hostname}:6379/0")
accept_content = ["moojson", "json"]
event_serializer = "moojson"
task_serializer = "moojson"
result_serializer = "moojson"
result_backend = "django-db"
cache_backend = "default"
beat_scheduler = "django_celery_beat.schedulers:DatabaseScheduler"
task_time_limit = 3
broker_connection_retry_on_startup = True

logging = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "simple": {
            "format": "%(asctime)s: %(levelname)s %(message)s",
        },
        "celeryTask": {
            "()": "celery.app.log.TaskFormatter",
            "fmt": "%(asctime)s: %(levelname)s %(task_name)s[%(task_id)s]: %(message)s",
        },
        "celeryProcess": {"()": "celery.utils.log.ColorFormatter", "fmt": "%(asctime)s: %(levelname)s %(message)s"},
    },
    "filters": {
        "celeryTask": {
            "()": "moo.logging.CeleryTaskFilter",
        },
        "celeryProcess": {
            "()": "moo.logging.CeleryProcessFilter",
        },
        "notCelery": {
            "()": "moo.logging.NotCeleryFilter",
        },
    },
    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "simple",
            "filters": ["notCelery"],
        },
        "console2": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "celeryTask",
            "filters": ["celeryTask"],
        },
        "console3": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "celeryProcess",
            "filters": ["celeryProcess"],
        },
    },
    "loggers": {
        "": {
            "handlers": ["console", "console2", "console3"],
            "level": "INFO",
            "propagate": False,
        }
    },
}



#!moo verb titlec --on $root_class

# pylint: disable=return-outside-function,undefined-variable

"""
This verb performs the same function as the `title` verb, but returns a capitalised version of the name property of the
object, using the string.capitalize() method.

"""

if not this.name:
    return str(this)
return this.name.capitalize()



#!moo verb enterfunc --on $room

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is invoked by the LambdaMOO server when an object moves into a room, as part of action of the `move` primitive.
The actions taken for a room are very straightforward. If `thing` is a player object, then we tell the player where s/he
has moved into using the `$room.look_self` verb on the room. If the object is the `blessed_object` for this room, then the
`blessed_object` property for the room is reset to None. For further details on blessed objects, refer to the
description of `$room.bless_for_entry`.
"""

from moo.core import context

thing = args[0]
if thing.is_player():
    thing.tell(this.look_self())

if thing == this.blessed_object and context.task_id == this.blessed_task_id:
    this.blessed_object = None
    this.blessed_task_id = None
    this.save()



#!moo verb @unlock --on $player --dspec any

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is a player command used to unlock an object. The direct object string is matched to try and find an object
to unlock. If a match is found, the `key` property is reset to `None'. Any errors are reported to the invoking player.
"""

from moo.core import context

parser = context.parser

obj = parser.get_dobj()
obj.set_property("key", None)



# -*- coding: utf-8 -*-
"""
Authentication support.
"""

from django.contrib.auth.models import \
    User  # pylint: disable=imported-auth-user
from django.db import models


class Player(models.Model):
    user = models.OneToOneField(User, null=True, on_delete=models.CASCADE)
    avatar = models.ForeignKey("Object", null=True, on_delete=models.SET_NULL)
    wizard = models.BooleanField(default=False)



#!moo verb @whereis whereis --on $player --dspec any

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is a player command used to locate a player in the virtual world. The code, shown below, allows any player to
locate any other player. If no argument is given, a list of the locations of all the currently connected players is
printed.

If an argument is given, the verb attempts to match with one or more player names. If no valid matches are found, a
suitable error message is printed by invoking the verb $command_utils:player_match_result. That verb returns a list.
The first element indicates whether any of the elements of the argument list didn't match. The rest of the list
contains the objects references that did match.

The verb runs through the list of object references, and reports the string returned by each player's whereis_location_msg verb.
"""

from moo.core import context, lookup

targets = args if args else [context.parser.get_dobj_str()]
matched = []
unmatched = []
for target in targets:
    try:
        result = lookup(target)
        matched.append(result)
    except type(this).NotFoundError:
        unmatched.append(target)

print("Found the following players:")
for player in matched:
    print(player.whereis_location_msg())



#!moo verb look --on $exit --dspec none --ispec through:any

# pylint: disable=return-outside-function,undefined-variable,redefined-builtin

from moo.core import context, lookup

door_description = context.parser.get_pobj_str("through")
exit = this

obj = exit.get_property("dest")
if not obj:
    print(f"There is no door called {door_description} here.")
    return

obj.look_self()



# -*- coding: utf-8 -*-
"""
Access control list functionality for Object, Verbs and Properties.
"""

import logging

from django.db import models

from .. import code

log = logging.getLogger(__name__)


class AccessibleMixin:
    """
    The base class for all Objects, Verbs, and Properties.
    """

    def can_caller(self, permission, subject):
        """
        Check if the current caller has permission for something.
        """
        caller = code.ContextManager.get("caller")
        if not caller:
            return
        if permission == "grant" and caller.owns(subject):
            return
        caller.is_allowed(permission, subject, fatal=True)

    def allow(self, accessor, permission):
        """
        Allow a certain object or group to do something on this object.

        [ACL] allowed to grant on this (or owner of this)
        """
        self.can_caller("grant", self)
        Access.objects.create(
            object=self if self.kind == "object" else None,
            verb=self if self.kind == "verb" else None,
            property=self if self.kind == "property" else None,
            rule="allow",
            permission=Permission.objects.get(name=permission),
            type="group" if isinstance(accessor, str) else "accessor",
            accessor=None if isinstance(accessor, str) else accessor,
            group=accessor if isinstance(accessor, str) else None,
        )

    def deny(self, accessor, permission):
        """
        Deny a certain object or group from doing something on this object.

        [ACL] allowed to grant on this (or owner of this)
        """
        self.can_caller("grant", self)
        Access.objects.create(
            object=self if self.kind == "object" else None,
            verb=self if self.kind == "verb" else None,
            property=self if self.kind == "property" else None,
            rule="deny",
            permission=Permission.objects.get(name=permission),
            type="group" if isinstance(accessor, str) else "accessor",
            accessor=None if isinstance(accessor, str) else accessor,
            group=accessor if isinstance(accessor, str) else None,
        )


class Permission(models.Model):
    name = models.CharField(max_length=255)

    def __str__(self):  # pylint: disable=invalid-str-returned
        return self.name


class Access(models.Model):
    class Meta:
        verbose_name_plural = "access controls"
        unique_together = ("object", "verb", "property", "rule", "permission", "type", "accessor", "group", "weight")

    object = models.ForeignKey("Object", related_name="acl", null=True, on_delete=models.CASCADE)
    verb = models.ForeignKey("Verb", related_name="acl", null=True, on_delete=models.CASCADE)
    property = models.ForeignKey("Property", related_name="acl", null=True, on_delete=models.CASCADE)
    rule = models.CharField(max_length=5, choices=[(x, x) for x in ("allow", "deny")])
    permission = models.ForeignKey(Permission, related_name="usage", on_delete=models.CASCADE)
    type = models.CharField(max_length=8, choices=[(x, x) for x in ("accessor", "group")])
    accessor = models.ForeignKey("Object", related_name="rights", null=True, on_delete=models.CASCADE)
    group = models.CharField(max_length=8, null=True, choices=[(x, x) for x in ("everyone", "owners", "wizards")])
    weight = models.IntegerField(default=0)

    def actor(self):
        return self.accessor if self.type == "accessor" else self.group

    def entity(self):
        if self.object:
            return "self"
        elif self.verb:
            return "".join(
                [
                    ["", "@"][self.verb.ability],  # pylint: disable=no-member
                    self.verb.names.all()[:1][0].name,
                    ["", "()"][self.verb.method],  # pylint: disable=no-member
                ]
            )
        else:
            return self.property.name

    def origin(self):
        if self.object:
            return self.object
        elif self.verb:
            return self.verb.origin  # pylint: disable=no-member
        else:
            return self.property.origin  # pylint: disable=no-member

    def __str__(self):
        return "%(rule)s %(actor)s %(permission)s on %(entity)s (%(weight)s)" % dict(
            rule=self.rule,
            actor=self.actor(),
            permission=self.permission.name,
            entity=self.entity(),
            weight=self.weight,
        )



#!moo verb describe --on $room --dspec this

# pylint: disable=return-outside-function,undefined-variable

from moo.core import lookup, context

obj = this

response = ""

if obj.has_property("description"):
    response += obj.get_property('description')
else:
    response += "[color deep_sky_blue1]No description available.[/color deep_sky_blue1]"

contents = obj.contents.filter(obvious=True)
if contents:
    response += "\n[color yellow]Obvious contents:[color /yellow]\n"
    for content in contents:
        response += f"{content.name}\n"

return response



#!moo verb test-args-parser --on "player class" --dspec none

# pylint: disable=return-outside-function,undefined-variable

from moo.core import context

if context.parser is not None:
    print("PARSER")



# Generated by Django 5.0 on 2024-02-14 01:59

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0007_property_inherited"),
    ]

    operations = [
        migrations.AlterField(
            model_name="relationship",
            name="child",
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name="+", to="core.object"),
        ),
        migrations.AlterField(
            model_name="relationship",
            name="parent",
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name="+", to="core.object"),
        ),
    ]



import pytest

from moo.core import code, create, lookup
from moo.core.models import Object


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_set_neuter(t_init: Object, t_wizard: Object):
    """Setting gender to 'neuter' assigns neuter pronouns to the object."""
    with code.ContextManager(t_wizard, lambda msg: None):
        system = lookup(1)
        obj = create("test_neuter", parents=[system.thing])
        result = system.gender_utils.set(obj, "neuter")
    assert result == "neuter"
    assert obj.get_property("ps") == "it"
    assert obj.get_property("po") == "it"
    assert obj.get_property("pp") == "its"
    assert obj.get_property("pr") == "itself"


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_set_male(t_init: Object, t_wizard: Object):
    """Setting gender to 'male' assigns masculine pronouns to the object."""
    with code.ContextManager(t_wizard, lambda msg: None):
        system = lookup(1)
        obj = create("test_male", parents=[system.thing])
        result = system.gender_utils.set(obj, "male")
    assert result == "male"
    assert obj.get_property("ps") == "he"
    assert obj.get_property("po") == "him"
    assert obj.get_property("pp") == "his"
    assert obj.get_property("pr") == "himself"


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_set_female(t_init: Object, t_wizard: Object):
    """Setting gender to 'female' assigns feminine pronouns to the object."""
    with code.ContextManager(t_wizard, lambda msg: None):
        system = lookup(1)
        obj = create("test_female", parents=[system.thing])
        result = system.gender_utils.set(obj, "female")
    assert result == "female"
    assert obj.get_property("ps") == "she"
    assert obj.get_property("po") == "her"
    assert obj.get_property("pp") == "her"
    assert obj.get_property("pr") == "herself"


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_set_plural(t_init: Object, t_wizard: Object):
    """Setting gender to 'plural' assigns plural pronouns to the object."""
    with code.ContextManager(t_wizard, lambda msg: None):
        system = lookup(1)
        obj = create("test_plural", parents=[system.thing])
        result = system.gender_utils.set(obj, "plural")
    assert result == "plural"
    assert obj.get_property("ps") == "they"
    assert obj.get_property("po") == "them"
    assert obj.get_property("pp") == "their"
    assert obj.get_property("pr") == "themselves"


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_set_unknown_gender_returns_none(t_init: Object, t_wizard: Object):
    """An unrecognised gender string causes set() to return None."""
    with code.ContextManager(t_wizard, lambda msg: None):
        system = lookup(1)
        obj = create("test_unknown", parents=[system.thing])
        result = system.gender_utils.set(obj, "adkjfhadslfjkg")
    assert result is None


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_set_all_pronoun_properties(t_init: Object, t_wizard: Object):
    """set() assigns all ten pronoun properties, including capitalised variants."""
    with code.ContextManager(t_wizard, lambda msg: None):
        system = lookup(1)
        obj = create("test_all_pronouns", parents=[system.thing])
        system.gender_utils.set(obj, "female")
    assert obj.get_property("ps") == "she"
    assert obj.get_property("po") == "her"
    assert obj.get_property("pp") == "her"
    assert obj.get_property("pr") == "herself"
    assert obj.get_property("pq") == "hers"
    assert obj.get_property("psc") == "She"
    assert obj.get_property("poc") == "Her"
    assert obj.get_property("ppc") == "Hers"
    assert obj.get_property("prc") == "Herself"
    assert obj.get_property("pqc") == "Hers"


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_set_overwrites_existing_pronouns(t_init: Object, t_wizard: Object):
    """set() replaces previously assigned pronouns when called a second time."""
    with code.ContextManager(t_wizard, lambda msg: None):
        system = lookup(1)
        obj = create("test_overwrite", parents=[system.thing])
        system.gender_utils.set(obj, "male")
        result = system.gender_utils.set(obj, "female")
    assert result == "female"
    assert obj.get_property("ps") == "she"
    assert obj.get_property("pr") == "herself"



#!moo verb @s*how --on $player --dspec any

# pylint: disable=return-outside-function,undefined-variable

"""
This is a player command used to examine other objects in detail. It returns details of an object's `owner`,
`location`, `parents`, and all the verbs and properties defined for the object, along with their values,
if permissions allow.
"""

from moo.core import context

parser = context.parser
player = context.player
obj = parser.get_dobj()

print(f"Details for {obj}:")
print(f"  Owner: {obj.owner}")
print(f"  Location: {obj.location}")
print(f"  Parents: {', '.join(str(p) for p in obj.parents.all())}")
print("  Verbs:")
for verb in obj.verbs.all():
    if player.is_allowed("execute", verb):
        print(f"    {', '.join(sorted(verb.names))}")
print("  Properties:")
for prop in obj.properties.all():
    if player.is_allowed("read", prop):
        print(f"    {prop.name}: {prop.value}")
